;; Generated package description from parser-generator.el  -*- no-byte-compile: t -*-
(define-package "parser-generator" "0.1.6.0.20220512.173154" "Parser Generator library" '((emacs "26")) :commit "0d0710d02ad002d52ba2f09f5c0a1cba85d83419" :authors '(("Christian Johansson" . "christian@cvj.se")) :maintainer '("Christian Johansson" . "christian@cvj.se") :keywords '("tools" "convenience") :url "https://github.com/cjohansson/emacs-parser-generator")
